program project1;

{$mode objfpc}{$H+}
{$warn 5024 off}

uses
  mycustfcgi, Classes, sysutils, custweb, HTTPDefs;

type
  TFastCGIRouter = class;

  { TFastCGIThread }

  TFastCGIThread = class(TThread)
  private
    FHandle: THandle;
    FKeepConnection: Boolean;
    FRouter: TFastCGIRouter;
  public
    constructor Create(const AHandle: THandle;
      const AKeepConnection: Boolean); virtual;
    destructor Destroy; override;
    procedure Execute; override;
    property KeepConnection: Boolean read FKeepConnection write FKeepConnection;
  end;

  { TFastCGIHandler }

  TFastCGIHandler = class(TFCgiHandler)
  private
    FKeepConnection: Boolean;
  public
    constructor Create(AOwner: TComponent;
      const AKeepConnection: Boolean); reintroduce; virtual;
    function WaitForRequest(out ARequest: TRequest;
      out AResponse: TResponse): Boolean; override;
    property KeepConnection: Boolean read FKeepConnection write FKeepConnection;
  end;

  { TFastCGIRouter }

  TFastCGIRouter = class(TFCgiHandler)
  private
    FKeepConnection: Boolean;
  public
    function WaitForRequest(out ARequest: TRequest;
      out AResponse: TResponse): Boolean; override;
    procedure HandleRequest(ARequest: TRequest; AResponse: TResponse); override;
    property Handle;
    property KeepConnection: Boolean read FKeepConnection write FKeepConnection;
  end;

  { TFastCGIApp }

  TFastCGIApp = class(TCustomFCgiApplication)
  private
    FKeepConnection: Boolean;
  public
    function InitializeWebHandler: TWebHandler; override;
    property KeepConnection: Boolean read FKeepConnection write FKeepConnection;
  end;

  { TFastCGIThread }

    constructor TFastCGIThread.Create(const AHandle: THandle;
    const AKeepConnection: Boolean);
  begin
    inherited Create(True);
    FreeOnTerminate := True;
    FKeepConnection := AKeepConnection;
    FHandle := AHandle;
    FRouter := TFastCGIRouter.Create(nil);
    FRouter.Handle := AHandle;
    FRouter.KeepConnection := FKeepConnection;
  end;

  destructor TFastCGIThread.Destroy;
  begin
    FRouter.Free;
    inherited Destroy;
  end;

  procedure TFastCGIThread.Execute;
  var
    VRequest: TRequest;
    VResponse: TResponse;
  begin

    while FRouter.WaitForRequest(VRequest, VResponse) do
      FRouter.DoHandleRequest(VRequest, VResponse);
  end;

  { TFastCGIHandler }

  constructor TFastCGIHandler.Create(AOwner: TComponent;
    const AKeepConnection: Boolean);
  begin
    inherited Create(AOwner);
    FKeepConnection := AKeepConnection;
  end;

  function TFastCGIHandler.WaitForRequest(out ARequest: TRequest;
    out AResponse: TResponse): Boolean;
  var
    VSocket: LongInt = 0;
  begin
    Result := False;
    if VSocket = 0 then
      if Port <> 0 then
        SetupSocket(FIAddress, FAddressLength)
      else
        VSocket := StdInputHandle;
    repeat
      VSocket := AcceptConnection;
      TFastCGIThread.Create(VSocket, FKeepConnection).Start;
      // we will implement the thread pool here ...
    until Terminated;
  end;

{ TFastCGIRouter }

  function TFastCGIRouter.WaitForRequest(out ARequest: TRequest;
    out AResponse: TResponse): Boolean;
  begin
    Result := inherited WaitForRequest(ARequest, AResponse);
  end;

  procedure TFastCGIRouter.HandleRequest(ARequest: TRequest;
    AResponse: TResponse);
  var
    I: Integer;
  begin
    if ARequest.QueryString = 'op=loop' then
    begin
      for I := 1 to 10 do
        Sleep(1000);
      AResponse.Content := 'Done.';
    end
    else
      AResponse.Content := 'Now: ' + FormatDateTime('hh:nn:ss.zzz', Now);
  end;

  { TFastCGIApp }

  function TFastCGIApp.InitializeWebHandler: TWebHandler;
  begin
    Result := TFastCGIHandler.Create(Self, FKeepConnection);
  end;

begin
  with TFastCGIApp.Create(nil) do
  try
    KeepConnection := True;
    Port := 8080;
    Run;
  finally
    Free;
  end;
end.

